import logging
import os
import shutil
from pathlib import Path

from lightning.pytorch import Trainer

from ..util.path import try_symlink_or_copy
from .metadata import _link_checkpoint_metadata, _remove_checkpoint_metadata

log = logging.getLogger(__name__)


def _link_checkpoint(
    filepath: str | Path | os.PathLike,
    linkpath: str | Path | os.PathLike,
    *,
    metadata: bool,
    remove_existing: bool = True,
):
    filepath = Path(filepath)
    linkpath = Path(linkpath)

    if remove_existing:
        try:
            if linkpath.exists():
                if linkpath.is_dir():
                    shutil.rmtree(linkpath)
                else:
                    linkpath.unlink(missing_ok=True)
        except Exception:
            log.warning(f"Failed to remove {linkpath}", exc_info=True)
        else:
            log.debug(f"Removed {linkpath=}")

        if metadata:
            _remove_checkpoint_metadata(linkpath)

    try_symlink_or_copy(filepath, linkpath)
    if metadata:
        _link_checkpoint_metadata(filepath, linkpath)


def _remove_checkpoint(
    trainer: Trainer,
    filepath: str | Path | os.PathLike,
    *,
    metadata: bool,
):
    filepath = Path(filepath)

    trainer.strategy.remove_checkpoint(filepath)

    if metadata:
        _remove_checkpoint_metadata(filepath)



from nshutils.actsave import *  # type: ignore # noqa: F403

from nshtrainer.callbacks.actsave import ActSaveCallback as ActSaveCallback
from nshtrainer.callbacks.actsave import ActSaveConfig as ActSaveConfig



"""Credit to useful-types from https://github.com/hauntsaninja/useful_types"""

from __future__ import annotations

from collections.abc import Awaitable, Iterable, Iterator, Sequence, Sized
from collections.abc import Set as AbstractSet
from os import PathLike
from typing import Any, TypeVar, overload

from typing_extensions import Buffer, Literal, Protocol, SupportsIndex, TypeAlias

_KT = TypeVar("_KT")
_KT_co = TypeVar("_KT_co", covariant=True)
_KT_contra = TypeVar("_KT_contra", contravariant=True)
_VT = TypeVar("_VT")
_VT_co = TypeVar("_VT_co", covariant=True)
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)

# For partially known annotations. Usually, fields where type annotations
# haven't been added are left unannotated, but in some situations this
# isn't possible or a type is already partially known. In cases like these,
# use Incomplete instead of Any as a marker. For example, use
# "Incomplete | None" instead of "Any | None".
Incomplete: TypeAlias = Any


class IdentityFunction(Protocol):
    def __call__(self, __x: _T) -> _T: ...


# ====================
# Comparison protocols
# ====================


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, __other: _T_contra) -> bool: ...


class SupportsDunderGT(Protocol[_T_contra]):
    def __gt__(self, __other: _T_contra) -> bool: ...


class SupportsDunderLE(Protocol[_T_contra]):
    def __le__(self, __other: _T_contra) -> bool: ...


class SupportsDunderGE(Protocol[_T_contra]):
    def __ge__(self, __other: _T_contra) -> bool: ...


class SupportsAllComparisons(
    SupportsDunderLT[Any],
    SupportsDunderGT[Any],
    SupportsDunderLE[Any],
    SupportsDunderGE[Any],
    Protocol,
): ...


SupportsRichComparison: TypeAlias = SupportsDunderLT[Any] | SupportsDunderGT[Any]
SupportsRichComparisonT = TypeVar(
    "SupportsRichComparisonT", bound=SupportsRichComparison
)

# ====================
# Dunder protocols
# ====================


class SupportsNext(Protocol[_T_co]):
    def __next__(self) -> _T_co: ...


class SupportsAnext(Protocol[_T_co]):
    def __anext__(self) -> Awaitable[_T_co]: ...


class SupportsAdd(Protocol[_T_contra, _T_co]):
    def __add__(self, __x: _T_contra) -> _T_co: ...


class SupportsRAdd(Protocol[_T_contra, _T_co]):
    def __radd__(self, __x: _T_contra) -> _T_co: ...


class SupportsSub(Protocol[_T_contra, _T_co]):
    def __sub__(self, __x: _T_contra) -> _T_co: ...


class SupportsRSub(Protocol[_T_contra, _T_co]):
    def __rsub__(self, __x: _T_contra) -> _T_co: ...


class SupportsDivMod(Protocol[_T_contra, _T_co]):
    def __divmod__(self, __other: _T_contra) -> _T_co: ...


class SupportsRDivMod(Protocol[_T_contra, _T_co]):
    def __rdivmod__(self, __other: _T_contra) -> _T_co: ...


# This protocol is generic over the iterator type, while Iterable is
# generic over the type that is iterated over.
class SupportsIter(Protocol[_T_co]):
    def __iter__(self) -> _T_co: ...


# This protocol is generic over the iterator type, while AsyncIterable is
# generic over the type that is iterated over.
class SupportsAiter(Protocol[_T_co]):
    def __aiter__(self) -> _T_co: ...


class SupportsLenAndGetItem(Protocol[_T_co]):
    def __len__(self) -> int: ...
    def __getitem__(self, __k: int) -> _T_co: ...


class SupportsTrunc(Protocol):
    def __trunc__(self) -> int: ...


# ====================
# Mapping-like protocols
# ====================


class SupportsItems(Protocol[_KT_co, _VT_co]):
    def items(self) -> AbstractSet[tuple[_KT_co, _VT_co]]: ...


class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, __key: _KT) -> _VT_co: ...


class SupportsGetItem(Protocol[_KT_contra, _VT_co]):
    def __contains__(self, __x: Any) -> bool: ...
    def __getitem__(self, __key: _KT_contra) -> _VT_co: ...


class SupportsItemAccess(SupportsGetItem[_KT_contra, _VT], Protocol[_KT_contra, _VT]):
    def __setitem__(self, __key: _KT_contra, __value: _VT) -> None: ...
    def __delitem__(self, __key: _KT_contra) -> None: ...


# ====================
# File handling
# ====================

StrPath: TypeAlias = str | PathLike[str]
BytesPath: TypeAlias = bytes | PathLike[bytes]
StrOrBytesPath: TypeAlias = str | bytes | PathLike[str] | PathLike[bytes]

OpenTextModeUpdating: TypeAlias = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
OpenTextModeWriting: TypeAlias = Literal[
    "w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"
]
OpenTextModeReading: TypeAlias = Literal[
    "r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"
]
OpenTextMode: TypeAlias = (
    OpenTextModeUpdating | OpenTextModeWriting | OpenTextModeReading
)
OpenBinaryModeUpdating: TypeAlias = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
OpenBinaryModeWriting: TypeAlias = Literal["wb", "bw", "ab", "ba", "xb", "bx"]
OpenBinaryModeReading: TypeAlias = Literal[
    "rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"
]
OpenBinaryMode: TypeAlias = (
    OpenBinaryModeUpdating | OpenBinaryModeReading | OpenBinaryModeWriting
)


class HasFileno(Protocol):
    def fileno(self) -> int: ...


FileDescriptor: TypeAlias = int
FileDescriptorLike: TypeAlias = int | HasFileno
FileDescriptorOrPath: TypeAlias = int | StrOrBytesPath


class SupportsRead(Protocol[_T_co]):
    def read(self, __length: int = ...) -> _T_co: ...


class SupportsReadline(Protocol[_T_co]):
    def readline(self, __length: int = ...) -> _T_co: ...


class SupportsNoArgReadline(Protocol[_T_co]):
    def readline(self) -> _T_co: ...


class SupportsWrite(Protocol[_T_contra]):
    def write(self, __s: _T_contra) -> object: ...


# ====================
# Buffer protocols
# ====================

# Unfortunately PEP 688 does not allow us to distinguish read-only
# from writable buffers. We use these aliases for readability for now.
# Perhaps a future extension of the buffer protocol will allow us to
# distinguish these cases in the type system.
ReadOnlyBuffer: TypeAlias = Buffer
# Anything that implements the read-write buffer interface.
WriteableBuffer: TypeAlias = Buffer
# Same as WriteableBuffer, but also includes read-only buffer types (like bytes).
ReadableBuffer: TypeAlias = Buffer


class SliceableBuffer(Buffer, Protocol):
    def __getitem__(self, __slice: slice) -> Sequence[int]: ...


class IndexableBuffer(Buffer, Protocol):
    def __getitem__(self, __i: int) -> int: ...


class SupportsGetItemBuffer(SliceableBuffer, IndexableBuffer, Protocol):
    def __contains__(self, __x: Any) -> bool: ...
    @overload
    def __getitem__(self, __slice: slice) -> Sequence[int]: ...
    @overload
    def __getitem__(self, __i: int) -> int: ...


class SizedBuffer(Sized, Buffer, Protocol): ...


# Source from https://github.com/python/typing/issues/256#issuecomment-1442633430
# This works because str.__contains__ does not accept object (either in typeshed or at runtime)
class SequenceNotStr(Protocol[_T_co]):
    @overload
    def __getitem__(self, index: SupportsIndex, /) -> _T_co: ...
    @overload
    def __getitem__(self, index: slice, /) -> Sequence[_T_co]: ...
    def __contains__(self, value: object, /) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_T_co]: ...
    def index(self, value: Any, start: int = 0, stop: int = ..., /) -> int: ...
    def count(self, value: Any, /) -> int: ...
    def __reversed__(self) -> Iterator[_T_co]: ...



import torch


def is_bf16_supported_no_emulation():
    r"""Return a bool indicating if the current CUDA/ROCm device supports dtype bfloat16."""
    version = getattr(torch, "version")

    # Check for ROCm, if true return true, no ROCM_VERSION check required,
    # since it is supported on AMD GPU archs.
    if version.hip:
        return True

    device = torch.cuda.current_device()

    # Check for CUDA version and device compute capability.
    # This is a fast way to check for it.
    cuda_version = version.cuda
    if (
        cuda_version is not None
        and int(cuda_version.split(".")[0]) >= 11
        and torch.cuda.get_device_properties(device).major >= 8
    ):
        return True

    return False



from typing import Annotated, TypeAlias

import nshconfig as C

from ._base import BaseProfilerConfig as BaseProfilerConfig
from .advanced import AdvancedProfilerConfig as AdvancedProfilerConfig
from .pytorch import PyTorchProfilerConfig as PyTorchProfilerConfig
from .simple import SimpleProfilerConfig as SimpleProfilerConfig

ProfilerConfig: TypeAlias = Annotated[
    SimpleProfilerConfig | AdvancedProfilerConfig | PyTorchProfilerConfig,
    C.Field(discriminator="name"),
]



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

import nshconfig as C
from lightning.pytorch.profilers import Profiler

if TYPE_CHECKING:
    from ..model import BaseConfig

log = logging.getLogger(__name__)


class BaseProfilerConfig(C.Config, ABC):
    dirpath: str | Path | None = None
    """
    Directory path for the ``filename``. If ``dirpath`` is ``None`` but ``filename`` is present, the
        ``trainer.log_dir`` (from :class:`~lightning.pytorch.loggers.tensorboard.TensorBoardLogger`)
        will be used.
    """
    filename: str | None = None
    """
    If present, filename where the profiler results will be saved instead of printing to stdout.
        The ``.txt`` extension will be used automatically.
    """

    @abstractmethod
    def create_profiler(self, root_config: "BaseConfig") -> Profiler | None: ...



import logging
import os
from pathlib import Path
from typing import Literal

from lightning.pytorch import Callback
from typing_extensions import override

from .base import CallbackConfigBase

log = logging.getLogger(__name__)


def _create_symlink_to_nshrunner(base_dir: Path):
    # Resolve the current nshrunner session directory
    if not (session_dir := os.environ.get("NSHRUNNER_SESSION_DIR")):
        log.warning("NSHRUNNER_SESSION_DIR is not set. Skipping symlink creation.")
        return
    session_dir = Path(session_dir)
    if not session_dir.exists() or not session_dir.is_dir():
        log.warning(
            f"NSHRUNNER_SESSION_DIR is not a valid directory: {session_dir}. "
            "Skipping symlink creation."
        )
        return

    # Create the symlink
    symlink_path = base_dir / "nshrunner"
    if symlink_path.exists():
        # If it already points to the correct directory, we're done
        if symlink_path.resolve() == session_dir.resolve():
            return

        # Otherwise, we should log a warning and remove the existing symlink
        log.warning(
            f"A symlink pointing to {symlink_path.resolve()} already exists at {symlink_path}. "
            "Removing the existing symlink."
        )
        symlink_path.unlink()

    symlink_path.symlink_to(session_dir)


class DirectorySetupConfig(CallbackConfigBase):
    name: Literal["directory_setup"] = "directory_setup"

    enabled: bool = True
    """Whether to enable the directory setup callback."""

    create_symlink_to_nshrunner_root: bool = True
    """Should we create a symlink to the root folder for the Runner (if we're in one)?"""

    def __bool__(self):
        return self.enabled

    def create_callbacks(self, root_config):
        if not self:
            return

        yield DirectorySetupCallback(self)


class DirectorySetupCallback(Callback):
    @override
    def __init__(self, config: DirectorySetupConfig):
        super().__init__()

        self.config = config
        del config

    @override
    def setup(self, trainer, pl_module, stage):
        super().setup(trainer, pl_module, stage)

        # Create a symlink to the root folder for the Runner
        if self.config.create_symlink_to_nshrunner_root:
            # Resolve the base dir
            from ..model.config import BaseConfig

            assert isinstance(
                config := pl_module.hparams, BaseConfig
            ), f"Expected a BaseConfig, got {type(config)}"

            base_dir = config.directory.resolve_run_root_directory(config.id)
            _create_symlink_to_nshrunner(base_dir)



import logging
import os
from contextlib import contextmanager

log = logging.getLogger(__name__)


@contextmanager
def remove_slurm_environment_variables():
    """
    SLURM_CPU_BIND_* environment variables are set by SLURM in the current environment.
    We need to remove all of these environment variables during the codepath in which we create the new SLURM runs, so that the new SLURM runs do not inherit the environment variables from the current environment.
    To make things easier, we will patch the environment to remove all "SLURM_" environment variables.
    Otherwise, the runs will faill with an error like shown below:
        srun: error: CPU binding outside of job step allocation, allocated CPUs are: 0x01F000000001F0000000.
        srun: error: Task launch for StepId=5216715.0 failed on node learnfair0537: Unable to satisfy cpu bind request
        srun: error: Application launch failed: Unable to satisfy cpu bind request
        srun: Job step aborted

    See https://www.mail-archive.com/slurm-users@lists.schedmd.com/msg09157.html for more details.
    """

    removed_env_vars = {}
    for key in list(os.environ.keys()):
        if not key.startswith("SLURM_"):
            continue
        removed_env_vars[key] = os.environ.pop(key)

    log.debug(
        f"Removed environment variables before launching new SLURM job: {list(removed_env_vars.keys())}"
    )
    try:
        yield
    finally:
        os.environ.update(removed_env_vars)
        log.debug(
            f"Restored environment variables after launching new SLURM job: {list(removed_env_vars.keys())}"
        )


@contextmanager
def remove_lsf_environment_variables():
    """Same as above, but for the LSF task scheduler."""
    removed_env_vars = {}
    for key in list(os.environ.keys()):
        if (
            not key.startswith("LS_")
            and not key.startswith("LSF_")
            and not key.startswith("LSB_")
            and not key.startswith("BSUB_")
        ):
            continue
        removed_env_vars[key] = os.environ.pop(key)

    log.debug(
        f"Removed environment variables before launching new LSF job: {list(removed_env_vars.keys())}"
    )
    try:
        yield
    finally:
        os.environ.update(removed_env_vars)
        log.debug(
            f"Restored environment variables after launching new LSF job: {list(removed_env_vars.keys())}"
        )


@contextmanager
def remove_wandb_environment_variables():
    """
    Similar to above, but removes all "WANDB_" environment variables.
    """

    removed_env_vars = {}
    for key in list(os.environ.keys()):
        if not key.startswith("WANDB_"):
            continue
        removed_env_vars[key] = os.environ.pop(key)

    log.debug(
        f"Removed environment variables before launching new SLURM job: {list(removed_env_vars.keys())}"
    )
    try:
        yield
    finally:
        os.environ.update(removed_env_vars)
        log.debug(
            f"Restored environment variables after launching new SLURM job: {list(removed_env_vars.keys())}"
        )


@contextmanager
def set_additional_env_vars(additional_env_vars: dict[str, str] | None = None):
    """
    Set additional environment variables for the run.
    Newly set environment variables will be removed after the run is finished.
    Existing environment variables will be restored to their original values after the run is finished.
    """
    if additional_env_vars is None:
        additional_env_vars = {}

    removed_env_vars = {}
    for key, value in additional_env_vars.items():
        removed_env_vars[key] = os.environ.pop(key, None)
        os.environ[key] = value

    log.debug(
        f"Set additional environment variables for the run: {list(additional_env_vars.keys())}"
    )
    try:
        yield
    finally:
        for key in additional_env_vars.keys():
            if removed_env_vars[key] is None:
                del os.environ[key]
            else:
                os.environ[key] = removed_env_vars[key]
        log.debug(
            f"Restored environment variables after launching new SLURM job: {list(additional_env_vars.keys())}"
        )



from nshtrainer.model import *  # noqa: F403

from ..trainer._config import CheckpointLoadingConfig as CheckpointLoadingConfig
from ..trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from ..trainer._config import GradientClippingConfig as GradientClippingConfig
from ..trainer._config import LoggingConfig as LoggingConfig
from ..trainer._config import OptimizationConfig as OptimizationConfig
from ..trainer._config import ReproducibilityConfig as ReproducibilityConfig
from ..trainer._config import SanityCheckingConfig as SanityCheckingConfig
from ..util._environment_info import (
    EnvironmentClassInformationConfig as EnvironmentClassInformationConfig,
)
from ..util._environment_info import EnvironmentConfig as EnvironmentConfig
from ..util._environment_info import (
    EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig,
)
from ..util._environment_info import (
    EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig,
)



import logging
from typing import Literal

from typing_extensions import override

from ._base import BaseProfilerConfig

log = logging.getLogger(__name__)


class AdvancedProfilerConfig(BaseProfilerConfig):
    name: Literal["advanced"] = "advanced"

    line_count_restriction: float = 1.0
    """
    This can be used to limit the number of functions
        reported for each action. either an integer (to select a count of lines),
        or a decimal fraction between 0.0 and 1.0 inclusive (to select a percentage of lines)
    """

    @override
    def create_profiler(self, root_config):
        from lightning.pytorch.profilers.advanced import AdvancedProfiler

        if (dirpath := self.dirpath) is None:
            dirpath = root_config.directory.resolve_subdirectory(
                root_config.id, "profile"
            )

        if (filename := self.filename) is None:
            filename = f"{root_config.id}_profile.txt"

        return AdvancedProfiler(
            line_count_restriction=self.line_count_restriction,
            dirpath=dirpath,
            filename=filename,
        )



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import nshconfig as C
from lightning.pytorch.loggers import Logger

if TYPE_CHECKING:
    from ..model import BaseConfig


class BaseLoggerConfig(C.Config, ABC):
    enabled: bool = True
    """Enable this logger."""

    priority: int = 0
    """Priority of the logger. Higher priority loggers are created first."""

    log_dir: C.DirectoryPath | None = None
    """Directory to save the logs to. If None, will use the default log directory for the trainer."""

    @abstractmethod
    def create_logger(self, root_config: "BaseConfig") -> Logger | None: ...

    def disable_(self):
        self.enabled = False
        return self



# type: ignore
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from collections import deque
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Deque,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
)

import torch
from lightning.fabric.plugins import Precision as FabricPrecision
from lightning.fabric.utilities.throughput import (
    _plugin_to_compute_dtype as fabric_plugin_to_compute_dtype,
)
from lightning.fabric.utilities.throughput import get_available_flops
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.plugins import (
    BitsandbytesPrecision,
    DeepSpeedPrecision,
    DoublePrecision,
    FSDPPrecision,
    HalfPrecision,
    MixedPrecision,
    Precision,
    TransformerEnginePrecision,
    XLAPrecision,
)
from lightning.pytorch.trainer.states import RunningStage, TrainerFn
from lightning.pytorch.utilities.rank_zero import rank_zero_only, rank_zero_warn
from typing_extensions import override

if TYPE_CHECKING:
    from lightning.pytorch import LightningModule, Trainer


_THROUGHPUT_METRICS = Dict[str, Union[int, float]]

T = TypeVar("T", bound=float)


class _MonotonicWindow(List[T]):
    """Custom fixed size list that only supports right-append and ensures that all values increase monotonically."""

    def __init__(self, maxlen: int) -> None:
        super().__init__()
        self.maxlen = maxlen

    @property
    def last(self) -> Optional[T]:
        if len(self) > 0:
            return self[-1]
        return None

    @override
    def append(self, x: T) -> None:
        last = self.last
        if last is not None and last >= x:
            rank_zero_warn(
                f"Expected the value to increase, last: {last}, current: {x}"
            )
        list.append(self, x)
        # truncate excess
        if len(self) > self.maxlen:
            del self[0]

    @override
    def __setitem__(self, key: Any, value: Any) -> None:
        # assigning is not implemented since we don't use it. it could be by checking all previous values
        raise NotImplementedError("__setitem__ is not supported")


class Throughput:
    """Computes throughput.

    +------------------------+-------------------------------------------------------------------------------------+
    | Key                    | Value                                                                               |
    +========================+=====================================================================================+
    | batches_per_sec        | Rolling average (over ``window_size`` most recent updates) of the number of batches |
    |                        | processed per second                                                                |
    +--------------------------+-----------------------------------------------------------------------------------+
    | samples_per_sec        | Rolling average (over ``window_size`` most recent updates) of the number of samples |
    |                        | processed per second                                                                |
    +--------------------------+-----------------------------------------------------------------------------------+
    | items_per_sec          | Rolling average (over ``window_size`` most recent updates) of the number of items   |
    |                        | processed per second                                                                |
    +--------------------------+-----------------------------------------------------------------------------------+
    | flpps_per_sec          | Rolling average (over ``window_size`` most recent updates) of the number of flops   |
    |                        | processed per second                                                                |
    +--------------------------+-----------------------------------------------------------------------------------+
    | device/batches_per_sec | batches_per_sec divided by world size                                               |
    +--------------------------+-----------------------------------------------------------------------------------+
    | device/samples_per_sec | samples_per_sec divided by world size                                               |
    +--------------------------+-----------------------------------------------------------------------------------+
    | device/items_per_sec   | items_per_sec divided by world size. This may include padding depending on the data |
    +--------------------------+-----------------------------------------------------------------------------------+
    | device/flops_per_sec   | flops_per_sec divided by world size.                                                |
    +--------------------------+-----------------------------------------------------------------------------------+
    | device/mfu             | device/flops_per_sec divided by world size.                                         |
    +--------------------------+-----------------------------------------------------------------------------------+
    | time                   | Total elapsed time                                                                  |
    +--------------------------+-----------------------------------------------------------------------------------+
    | batches                | Total batches seen                                                                  |
    +--------------------------+-----------------------------------------------------------------------------------+
    | samples                | Total samples seen                                                                  |
    +--------------------------+-----------------------------------------------------------------------------------+
    | lengths                | Total items seen                                                                    |
    +--------------------------+-----------------------------------------------------------------------------------+

    Example::

        throughput = Throughput()
        t0 = time()
        for i in range(1000):
            do_work()
            if torch.cuda.is_available(): torch.cuda.synchronize()  # required or else time() won't be correct
            throughput.update(time=time() - t0, samples=i)
            if i % 10 == 0:
                print(throughput.compute())

    Notes:
        - The implementation assumes that devices FLOPs are all the same as it normalizes by the world size and only
          takes a single ``available_flops`` value.
        - items_per_sec, flops_per_sec and MFU do not account for padding if present. We suggest using
          samples_per_sec or batches_per_sec to measure throughput under this circumstance.

    Args:
        available_flops: Number of theoretical flops available for a single device.
        world_size: Number of devices available across hosts. Global metrics are not included if the world size is 1.
        window_size: Number of batches to use for a rolling average.
        separator: Key separator to use when creating per-device and global metrics.

    """

    def __init__(
        self,
        available_flops: Optional[float] = None,
        world_size: int = 1,
        window_size: int = 100,
        separator: str = "/",
    ) -> None:
        self.available_flops = available_flops
        self.separator = separator
        assert world_size > 0
        self.world_size = world_size

        # throughput is computed over a window of values. at least 2 is enforced since it looks at the difference
        # between the first and last elements
        assert window_size > 1
        # custom class instead of `deque(maxlen=)` because it's easy for users to mess up their timer/counters and log
        # values that do not increase monotonically. this class will raise an error if that happens.
        self._time: _MonotonicWindow[float] = _MonotonicWindow(maxlen=window_size)
        self._batches: _MonotonicWindow[int] = _MonotonicWindow(maxlen=window_size)
        self._samples: _MonotonicWindow[int] = _MonotonicWindow(maxlen=window_size)
        self._lengths: _MonotonicWindow[int] = _MonotonicWindow(maxlen=window_size)
        self._flops: Deque[int] = deque(maxlen=window_size)

    def update(
        self,
        *,
        time: float,
        batches: int,
        samples: int,
        lengths: Optional[int] = None,
        flops: Optional[int] = None,
    ) -> None:
        """Update throughput metrics.

        Args:
            time: Total elapsed time in seconds. It should monotonically increase by the iteration time with each
                call.
            batches: Total batches seen per device. It should monotonically increase with each call.
            samples: Total samples seen per device. It should monotonically increase by the batch size with each call.
            lengths: Total length of the samples seen. It should monotonically increase by the lengths of a batch with
                each call.
            flops: Flops elapased per device since last ``update()`` call. You can easily compute this by using
                :func:`measure_flops` and multiplying it by the number of batches that have been processed.
                The value might be different in each device if the batch size is not the same.

        """
        self._time.append(time)
        if samples < batches:
            raise ValueError(
                f"Expected samples ({samples}) to be greater or equal than batches ({batches})"
            )
        self._batches.append(batches)
        self._samples.append(samples)
        if lengths is not None:
            if lengths < samples:
                raise ValueError(
                    f"Expected lengths ({lengths}) to be greater or equal than samples ({samples})"
                )
            self._lengths.append(lengths)
            if len(self._samples) != len(self._lengths):
                raise RuntimeError(
                    f"If lengths are passed ({len(self._lengths)}), there needs to be the same number of samples"
                    f" ({len(self._samples)})"
                )
        if flops is not None:
            # sum of flops across ranks
            self._flops.append(flops * self.world_size)

    def compute(self) -> _THROUGHPUT_METRICS:
        """Compute throughput metrics."""
        metrics = {
            "time": self._time[-1],
            "batches": self._batches[-1],
            "samples": self._samples[-1],
        }
        if self._lengths:
            metrics["lengths"] = self._lengths[-1]

        add_global_metrics = self.world_size > 1
        # a different but valid design choice would be to still compute all these metrics even if the window of values
        # has not been filled
        if len(self._time) == self._time.maxlen:
            elapsed_time = self._time[-1] - self._time[0]
            elapsed_batches = self._batches[-1] - self._batches[0]
            elapsed_samples = self._samples[-1] - self._samples[0]
            # we are safe from ZeroDivisionError thanks to `_MonotonicWindow`
            dev_samples_per_sec = elapsed_samples / elapsed_time
            dev_batches_per_sec = elapsed_batches / elapsed_time
            metrics.update(
                {
                    f"device{self.separator}batches_per_sec": elapsed_batches
                    / elapsed_time,
                    f"device{self.separator}samples_per_sec": dev_samples_per_sec,
                }
            )
            if add_global_metrics:
                samples_per_sec = dev_batches_per_sec * self.world_size
                metrics.update(
                    {
                        "batches_per_sec": samples_per_sec,
                        "samples_per_sec": dev_samples_per_sec * self.world_size,
                    }
                )

            if len(self._lengths) == self._lengths.maxlen:
                elapsed_lengths = self._lengths[-1] - self._lengths[0]
                dev_items_per_sec = elapsed_lengths / elapsed_time
                metrics[f"device{self.separator}items_per_sec"] = dev_items_per_sec
                if add_global_metrics:
                    items_per_sec = dev_items_per_sec * self.world_size
                    metrics["items_per_sec"] = items_per_sec

        if len(self._flops) == self._flops.maxlen:
            elapsed_flops = sum(self._flops) - self._flops[0]
            elapsed_time = self._time[-1] - self._time[0]
            flops_per_sec = elapsed_flops / elapsed_time
            dev_flops_per_sec = flops_per_sec / self.world_size
            if add_global_metrics:
                metrics["flops_per_sec"] = flops_per_sec
            metrics[f"device{self.separator}flops_per_sec"] = dev_flops_per_sec
            if self.available_flops:
                metrics[f"device{self.separator}mfu"] = (
                    dev_flops_per_sec / self.available_flops
                )

        return metrics

    def reset(self) -> None:
        self._time.clear()
        self._batches.clear()
        self._samples.clear()
        self._lengths.clear()
        self._flops.clear()


class ThroughputMonitor(Callback):
    r"""Computes and logs throughput with the :class:`~lightning.fabric.utilities.throughput.Throughput`

    Example::

        class MyModel(LightningModule):
            def setup(self, stage):
                with torch.device("meta"):
                    model = MyModel()

                    def sample_forward():
                        batch = torch.randn(..., device="meta")
                        return model(batch)

                    self.flops_per_batch = measure_flops(model, sample_forward, loss_fn=torch.Tensor.sum)


        logger = ...
        throughput = ThroughputMonitor(batch_size_fn=lambda batch: batch.size(0))
        trainer = Trainer(max_steps=1000, log_every_n_steps=10, callbacks=throughput, logger=logger)
        model = MyModel()
        trainer.fit(model)

    Notes:
        - It assumes that the batch size is the same during all iterations.
        - It will try to access a ``flops_per_batch`` attribute on your ``LightningModule`` on every iteration.
          We suggest using the :func:`~lightning.fabric.utilities.throughput.measure_flops` function for this.
          You might want to compute it differently each time based on your setup.

    Args:
        batch_size_fn: A function to compute the number of samples given a batch.
        length_fn: A function to compute the number of items in a sample given a batch.
        \**kwargs: See available parameters in
            :class:`~lightning.fabric.utilities.throughput.Throughput`

    """

    def __init__(
        self,
        batch_size_fn: Callable[[Any], int],
        length_fn: Optional[Callable[[Any], int | None]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.kwargs = kwargs
        self.batch_size_fn = batch_size_fn
        self.length_fn = length_fn
        self.available_flops: Optional[int] = None
        self._throughputs: Dict[RunningStage, Throughput] = {}
        self._t0s: Dict[RunningStage, float] = {}
        self._samples: Dict[RunningStage, int] = {}
        self._lengths: Dict[RunningStage, int] = {}

    @override
    def setup(
        self, trainer: "Trainer", pl_module: "LightningModule", stage: str
    ) -> None:
        dtype = _plugin_to_compute_dtype(trainer.precision_plugin)
        self.available_flops = get_available_flops(trainer.strategy.root_device, dtype)

        if stage == TrainerFn.FITTING and trainer.enable_validation:
            # `fit` includes validation inside
            throughput = Throughput(
                available_flops=self.available_flops,
                world_size=trainer.world_size,
                **self.kwargs,
            )
            self._throughputs[RunningStage.VALIDATING] = throughput

        throughput = Throughput(
            available_flops=self.available_flops,
            world_size=trainer.world_size,
            **self.kwargs,
        )
        stage = trainer.state.stage
        assert stage is not None
        self._throughputs[stage] = throughput

    def _start(self, trainer: "Trainer") -> None:
        stage = trainer.state.stage
        assert stage is not None
        self._throughputs[stage].reset()
        self._samples[stage] = 0
        self._lengths[stage] = 0
        self._t0s[stage] = time.perf_counter()

    def _update(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        batch: Any,
        iter_num: int,
    ) -> None:
        stage = trainer.state.stage
        assert stage is not None
        throughput = self._throughputs[stage]

        if trainer.strategy.root_device.type == "cuda":
            # required or else perf_counter() won't be correct
            torch.cuda.synchronize()

        elapsed = time.perf_counter() - self._t0s[stage]
        if self.length_fn is not None:
            with torch.inference_mode():
                if (length := self.length_fn(batch)) is not None:
                    self._lengths[stage] += length

        if hasattr(pl_module, "flops_per_batch"):
            flops_per_batch = pl_module.flops_per_batch
        else:
            rank_zero_warn(
                "When using the `ThroughputMonitor`, you need to define a `flops_per_batch` attribute or property"
                f" in {type(pl_module).__name__} to compute the FLOPs."
            )
            flops_per_batch = None

        with torch.inference_mode():
            self._samples[stage] += self.batch_size_fn(batch)

        throughput.update(
            time=elapsed,
            batches=iter_num,
            samples=self._samples[stage],
            lengths=None if self.length_fn is None else self._lengths[stage],
            flops=flops_per_batch,
        )

    def _compute(self, trainer: "Trainer", iter_num: Optional[int] = None) -> None:
        if not trainer._logger_connector.should_update_logs:
            return
        stage = trainer.state.stage
        assert stage is not None
        throughput = self._throughputs[stage]
        metrics = throughput.compute()
        # prefix with the stage to avoid collisions
        metrics = {
            f"{stage.value}{throughput.separator}{k}": v for k, v in metrics.items()
        }
        trainer._logger_connector.log_metrics(metrics, step=iter_num)  # type: ignore[arg-type]

    @override
    @rank_zero_only
    def on_train_start(self, trainer: "Trainer", *_: Any) -> None:
        self._start(trainer)

    @override
    @rank_zero_only
    def on_train_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: Any,
        batch: Any,
        *_: Any,
    ) -> None:
        self._update(trainer, pl_module, batch, trainer.fit_loop.total_batch_idx + 1)
        # log only when gradient accumulation is over. this ensures that we only measure when the effective batch has
        # finished and the `optimizer.step()` time is included
        if not trainer.fit_loop._should_accumulate():
            self._compute(trainer)

    @override
    @rank_zero_only
    def on_validation_start(self, trainer: "Trainer", *_: Any) -> None:
        if trainer.sanity_checking:
            return
        self._start(trainer)

    @override
    @rank_zero_only
    def on_validation_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: Any,
        batch: Any,
        *_: Any,
        **__: Any,
    ) -> None:
        if trainer.sanity_checking:
            return
        iter_num = trainer._evaluation_loop.batch_progress.total.ready
        self._update(trainer, pl_module, batch, iter_num)
        self._compute(trainer, iter_num)

    @override
    @rank_zero_only
    def on_validation_end(self, trainer: "Trainer", *_: Any) -> None:
        if trainer.sanity_checking or trainer.state.fn != TrainerFn.FITTING:
            return
        # add the validation time to the training time before continuing to avoid sinking the training throughput
        training_finished = self._t0s[RunningStage.TRAINING] + sum(
            self._throughputs[RunningStage.TRAINING]._time
        )
        time_between_train_and_val = (
            self._t0s[RunningStage.VALIDATING] - training_finished
        )
        val_time = sum(self._throughputs[RunningStage.VALIDATING]._time)
        self._t0s[RunningStage.TRAINING] += time_between_train_and_val + val_time

    @override
    @rank_zero_only
    def on_test_start(self, trainer: "Trainer", *_: Any) -> None:
        self._start(trainer)

    @override
    @rank_zero_only
    def on_test_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: Any,
        batch: Any,
        *_: Any,
        **__: Any,
    ) -> None:
        iter_num = trainer._evaluation_loop.batch_progress.total.ready
        self._update(trainer, pl_module, batch, iter_num)
        self._compute(trainer, iter_num)

    @override
    @rank_zero_only
    def on_predict_start(self, trainer: "Trainer", *_: Any) -> None:
        self._start(trainer)

    @override
    @rank_zero_only
    def on_predict_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: Any,
        batch: Any,
        *_: Any,
        **__: Any,
    ) -> None:
        iter_num = trainer.predict_loop.batch_progress.total.ready
        self._update(trainer, pl_module, batch, iter_num)
        self._compute(trainer, iter_num)


def _plugin_to_compute_dtype(plugin: Union[FabricPrecision, Precision]) -> torch.dtype:
    # TODO: integrate this into the precision plugins
    if not isinstance(plugin, Precision):
        return fabric_plugin_to_compute_dtype(plugin)
    if isinstance(plugin, BitsandbytesPrecision):
        return plugin.dtype
    if isinstance(plugin, HalfPrecision):
        return plugin._desired_input_dtype
    if isinstance(plugin, MixedPrecision):
        return torch.bfloat16 if plugin.precision == "bf16-mixed" else torch.half
    if isinstance(plugin, DoublePrecision):
        return torch.double
    if isinstance(plugin, (XLAPrecision, DeepSpeedPrecision)):
        return plugin._desired_dtype
    if isinstance(plugin, TransformerEnginePrecision):
        return torch.int8
    if isinstance(plugin, FSDPPrecision):
        return plugin.mixed_precision_config.reduce_dtype or torch.float32
    if isinstance(plugin, Precision):
        return torch.float32
    raise NotImplementedError(plugin)



from typing import TYPE_CHECKING, Literal, TypeAlias

import nshconfig as C
import torch
from typing_extensions import assert_never

from ..bf16 import is_bf16_supported_no_emulation

if TYPE_CHECKING:
    from ...model.base import BaseConfig

DTypeName: TypeAlias = Literal[
    "float32",
    "float",
    "float64",
    "double",
    "float16",
    "bfloat16",
    "float8_e4m3fn",
    "float8_e4m3fnuz",
    "float8_e5m2",
    "float8_e5m2fnuz",
    "half",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "short",
    "int32",
    "int",
    "int64",
    "long",
    "complex32",
    "complex64",
    "chalf",
    "cfloat",
    "complex128",
    "cdouble",
    "quint8",
    "qint8",
    "qint32",
    "bool",
    "quint4x2",
    "quint2x4",
    "bits1x8",
    "bits2x4",
    "bits4x2",
    "bits8",
    "bits16",
]


class DTypeConfig(C.Config):
    name: DTypeName
    """The name of the dtype."""

    @classmethod
    def from_base_config(cls, config: "BaseConfig"):
        if (precision := config.trainer.precision) is None:
            precision = "32-true"

        match precision:
            case "16-mixed-auto":
                return (
                    cls(name="bfloat16")
                    if is_bf16_supported_no_emulation()
                    else cls(name="float16")
                )
            case "fp16-mixed":
                return cls(name="float16")
            case "bf16-mixed":
                return cls(name="bfloat16")
            case "32-true":
                return cls(name="float32")
            case "64-true":
                return cls(name="float64")
            case _:
                assert_never(config.trainer.precision)

    @property
    def torch_dtype(self):
        if ((dtype := getattr(torch, self.name, None)) is None) or not isinstance(
            dtype, torch.dtype
        ):
            raise ValueError(f"Unknown dtype {self.name}")

        return dtype



import math
import warnings
from typing import Literal

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from typing_extensions import override

from ..util.config import DurationConfig
from ._base import LRSchedulerConfigBase, LRSchedulerMetadata


class LinearWarmupCosineAnnealingLR(LRScheduler):
    _get_lr_called_within_step: bool

    def __init__(
        self,
        optimizer: Optimizer,
        warmup_epochs: int,
        max_epochs: int,
        warmup_start_lr: float = 0.0,
        eta_min: float = 0.0,
        last_epoch: int = -1,
        should_restart: bool = True,
    ) -> None:
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs
        self.warmup_start_lr = warmup_start_lr
        self.eta_min = eta_min
        self.should_restart = should_restart

        super().__init__(optimizer, last_epoch)

    @override
    def get_lr(self) -> list[float]:  # pyright: ignore[reportIncompatibleMethodOverride]
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                UserWarning,
            )

        if self.last_epoch == 0:
            return [self.warmup_start_lr] * len(self.base_lrs)
        if self.last_epoch < self.warmup_epochs:
            return [
                group["lr"]
                + (base_lr - self.warmup_start_lr) / (self.warmup_epochs - 1)
                for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)
            ]
        if self.last_epoch == self.warmup_epochs:
            return self.base_lrs

        if not self.should_restart and self.last_epoch >= self.max_epochs:
            return [self.eta_min] * len(self.base_lrs)

        if (self.last_epoch - 1 - self.max_epochs) % (
            2 * (self.max_epochs - self.warmup_epochs)
        ) == 0:
            return [
                group["lr"]
                + (base_lr - self.eta_min)
                * (1 - math.cos(math.pi / (self.max_epochs - self.warmup_epochs)))
                / 2
                for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)
            ]

        return [
            (
                1
                + math.cos(
                    math.pi
                    * (self.last_epoch - self.warmup_epochs)
                    / (self.max_epochs - self.warmup_epochs)
                )
            )
            / (
                1
                + math.cos(
                    math.pi
                    * (self.last_epoch - self.warmup_epochs - 1)
                    / (self.max_epochs - self.warmup_epochs)
                )
            )
            * (group["lr"] - self.eta_min)
            + self.eta_min
            for group in self.optimizer.param_groups
        ]


class LinearWarmupCosineDecayLRSchedulerConfig(LRSchedulerConfigBase):
    name: Literal["linear_warmup_cosine_decay"] = "linear_warmup_cosine_decay"

    warmup_duration: DurationConfig
    r"""The duration for the linear warmup phase.
    The learning rate is linearly increased from `warmup_start_lr` to the initial learning rate over this duration."""

    max_duration: DurationConfig
    r"""The total duration.
    The learning rate is decayed to `min_lr` over this duration."""

    warmup_start_lr_factor: float = 0.0
    r"""The initial learning rate for the linear warmup phase, as a factor of the initial learning rate.
    The learning rate is linearly increased from this value to the initial learning rate over `warmup_epochs` epochs."""

    min_lr_factor: float = 0.0
    r"""The minimum learning rate, as a factor of the initial learning rate.
    The learning rate is decayed to this value over `max_epochs` epochs."""

    annealing: bool = False
    r"""Whether to restart the learning rate schedule after `max_epochs` epochs.
    If `False`, the learning rate will be decayed to `min_lr` over `max_epochs` epochs, and then the learning rate will be set to `min_lr` for all subsequent epochs.
    If `True`, the learning rate will be decayed to `min_lr` over `max_epochs` epochs, and then the learning rate will be increased back to the initial learning rate over `max_epochs` epochs, and so on (this is called a cosine annealing schedule)."""

    @override
    def metadata(self) -> LRSchedulerMetadata:
        return {
            "interval": "step",
        }

    @override
    def create_scheduler_impl(self, optimizer, lightning_module, lr):
        num_steps_per_epoch = self.compute_num_steps_per_epoch(lightning_module)
        warmup_steps = self.warmup_duration.to_steps(num_steps_per_epoch).value
        max_steps = self.max_duration.to_steps(num_steps_per_epoch).value
        warmup_start_lr = self.warmup_start_lr_factor * lr
        min_lr = self.min_lr_factor * lr

        # Warmup and max steps should be at least 1.
        warmup_steps = max(warmup_steps, 1)
        max_steps = max(max_steps, 1)

        # Create the scheduler
        scheduler = LinearWarmupCosineAnnealingLR(
            optimizer=optimizer,
            warmup_epochs=warmup_steps,
            max_epochs=max_steps,
            warmup_start_lr=warmup_start_lr,
            eta_min=min_lr,
            should_restart=self.annealing,
        )
        return scheduler



from pathlib import Path
from typing import TYPE_CHECKING

from lightning.pytorch.callbacks import Callback as _LightningCallback

if TYPE_CHECKING:
    from .model import LightningModuleBase
    from .trainer import Trainer


class NTCallbackBase(_LightningCallback):
    def on_checkpoint_saved(
        self,
        ckpt_path: Path,
        metadata_path: Path | None,
        trainer: "Trainer",
        pl_module: "LightningModuleBase",
    ) -> None:
        """Called after a checkpoint is saved."""
        pass


def _call_on_checkpoint_saved(
    trainer: "Trainer",
    ckpt_path: str | Path,
    metadata_path: str | Path | None,
):
    ckpt_path = Path(ckpt_path)
    metadata_path = Path(metadata_path) if metadata_path else None

    for callback in trainer.callbacks:
        if not isinstance(callback, NTCallbackBase):
            continue

        callback.on_checkpoint_saved(
            ckpt_path,
            metadata_path,
            trainer,
            trainer._base_module,
        )



import logging
from typing import TYPE_CHECKING, Literal, cast

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from typing_extensions import override

from .base import CallbackConfigBase

if TYPE_CHECKING:
    from ..model.config import BaseConfig

log = logging.getLogger(__name__)


class DebugFlagCallbackConfig(CallbackConfigBase):
    name: Literal["debug_flag"] = "debug_flag"

    enabled: bool = True
    """Whether to enable the callback."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, root_config):
        if not self:
            return

        yield DebugFlagCallback(self)


class DebugFlagCallback(Callback):
    """
    Sets the debug flag to true in the following circumstances:
    - fast_dev_run is enabled
    - sanity check is running
    """

    @override
    def __init__(self, config: DebugFlagCallbackConfig):
        super().__init__()

        self.config = config
        del config

    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str):
        if not getattr(trainer, "fast_dev_run", False):
            return

        hparams = cast("BaseConfig", pl_module.hparams)
        if not hparams.debug:
            log.critical("Fast dev run detected, setting debug flag to True.")
        hparams.debug = True

    @override
    def on_sanity_check_start(self, trainer: Trainer, pl_module: LightningModule):
        hparams = cast("BaseConfig", pl_module.hparams)
        self._debug = hparams.debug
        if not self._debug:
            log.critical("Enabling debug flag during sanity check routine.")
        hparams.debug = True

    @override
    def on_sanity_check_end(self, trainer: Trainer, pl_module: LightningModule):
        hparams = cast("BaseConfig", pl_module.hparams)
        if not self._debug:
            log.critical("Sanity check routine complete, disabling debug flag.")
        hparams.debug = self._debug



from abc import ABC, abstractmethod
from typing import Annotated, Literal

import nshconfig as C
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing_extensions import final, override


class BaseNonlinearityConfig(C.Config, ABC):
    @abstractmethod
    def create_module(self) -> nn.Module: ...

    @abstractmethod
    def __call__(self, x: torch.Tensor) -> torch.Tensor: ...


@final
class ReLUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["relu"] = "relu"

    @override
    def create_module(self) -> nn.Module:
        return nn.ReLU()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.relu(x)


@final
class SigmoidNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["sigmoid"] = "sigmoid"

    @override
    def create_module(self) -> nn.Module:
        return nn.Sigmoid()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sigmoid(x)


@final
class TanhNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["tanh"] = "tanh"

    @override
    def create_module(self) -> nn.Module:
        return nn.Tanh()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return torch.tanh(x)


@final
class SoftmaxNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["softmax"] = "softmax"

    dim: int = -1
    """The dimension to apply the softmax function."""

    @override
    def create_module(self) -> nn.Module:
        return nn.Softmax(dim=self.dim)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return torch.softmax(x, dim=self.dim)


@final
class SoftplusNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["softplus"] = "softplus"

    beta: float = 1.0
    """The beta parameter in the softplus function."""

    threshold: float = 20.0
    """Values above this revert to a linear function."""

    @override
    def create_module(self) -> nn.Module:
        return nn.Softplus(beta=self.beta, threshold=self.threshold)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.softplus(x, beta=self.beta, threshold=self.threshold)


@final
class SoftsignNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["softsign"] = "softsign"

    @override
    def create_module(self) -> nn.Module:
        return nn.Softsign()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.softsign(x)


@final
class ELUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["elu"] = "elu"

    alpha: float = 1.0
    """The alpha parameter in the ELU function."""

    @override
    def create_module(self) -> nn.Module:
        return nn.ELU()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.elu(x, alpha=self.alpha)


@final
class LeakyReLUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["leaky_relu"] = "leaky_relu"

    negative_slope: float = 1.0e-2
    """The negative slope of the leaky ReLU function."""

    @override
    def create_module(self) -> nn.Module:
        return nn.LeakyReLU(negative_slope=self.negative_slope)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.leaky_relu(x, negative_slope=self.negative_slope)


@final
class PReLUConfig(BaseNonlinearityConfig):
    name: Literal["prelu"] = "prelu"

    num_parameters: int = 1
    """The number of :math:`a` to learn.
    Although it takes an int as input, there is only two values are legitimate:
    1, or the number of channels at input."""

    init: float = 0.25
    """The initial value of :math:`a`."""

    @override
    def create_module(self) -> nn.Module:
        return nn.PReLU(num_parameters=self.num_parameters, init=self.init)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError(
            "PReLU requires learnable parameters and cannot be called directly."
        )


@final
class GELUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["gelu"] = "gelu"

    approximate: Literal["tanh", "none"] = "none"
    """The gelu approximation algorithm to use."""

    @override
    def create_module(self) -> nn.Module:
        return nn.GELU(approximate=self.approximate)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.gelu(x, approximate=self.approximate)


@final
class SwishNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["swish"] = "swish"

    @override
    def create_module(self) -> nn.Module:
        return nn.SiLU()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.silu(x)


@final
class SiLUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["silu"] = "silu"

    @override
    def create_module(self) -> nn.Module:
        return nn.SiLU()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.silu(x)


@final
class MishNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["mish"] = "mish"

    @override
    def create_module(self) -> nn.Module:
        return nn.Mish()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return F.mish(x)


class SwiGLU(nn.SiLU):
    @override
    def forward(self, input: torch.Tensor):
        input, gate = input.chunk(2, dim=-1)
        return input * super().forward(gate)


@final
class SwiGLUNonlinearityConfig(BaseNonlinearityConfig):
    name: Literal["swiglu"] = "swiglu"

    @override
    def create_module(self) -> nn.Module:
        return SwiGLU()

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        input, gate = x.chunk(2, dim=-1)
        return input * F.silu(gate)


NonlinearityConfig = Annotated[
    ReLUNonlinearityConfig
    | SigmoidNonlinearityConfig
    | TanhNonlinearityConfig
    | SoftmaxNonlinearityConfig
    | SoftplusNonlinearityConfig
    | SoftsignNonlinearityConfig
    | ELUNonlinearityConfig
    | LeakyReLUNonlinearityConfig
    | PReLUConfig
    | GELUNonlinearityConfig
    | SwishNonlinearityConfig
    | SiLUNonlinearityConfig
    | MishNonlinearityConfig
    | SwiGLUNonlinearityConfig,
    C.Field(discriminator="name"),
]



import logging
import os
import platform
import re
import signal
import subprocess
import sys
import threading
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from types import FrameType
from typing import Any, TypeAlias

import nshrunner as nr
import torch.utils.data
from lightning.fabric.plugins.environments.lsf import LSFEnvironment
from lightning.fabric.plugins.environments.slurm import SLURMEnvironment
from lightning.pytorch.trainer.connectors.signal_connector import _HandlersCompose
from lightning.pytorch.trainer.connectors.signal_connector import (
    _SignalConnector as _LightningSignalConnector,
)
from typing_extensions import override

log = logging.getLogger(__name__)

_SIGNUM = int | signal.Signals
_HANDLER: TypeAlias = Callable[[_SIGNUM, FrameType], Any] | int | signal.Handlers | None
_IS_WINDOWS = platform.system() == "Windows"


def _resolve_requeue_signals():
    if (session := nr.Session.from_current_session()) is None:
        return None

    signals: list[signal.Signals] = []
    if session.submit_timeout_signal:
        signals.append(session.submit_timeout_signal)
    if session.submit_preempt_signal:
        signals.append(session.submit_preempt_signal)
    return signals


class _SignalConnector(_LightningSignalConnector):
    def _auto_requeue_signals(self) -> list[signal.Signals] | None:
        if not (signals := _resolve_requeue_signals()):
            return None

        signals_set = set(signals)
        valid_signals: set[signal.Signals] = signal.valid_signals()
        assert signals_set.issubset(
            valid_signals
        ), f"Invalid signal(s) found: {signals_set - valid_signals}"
        return signals

    def _compose_and_register(
        self,
        signum: signal.Signals,
        handlers: list[_HANDLER],
        replace_existing: bool = False,
    ):
        if _IS_WINDOWS:
            log.info(
                f"Signal {signum.name} has no handlers or is not supported on Windows."
            )
            return

        if self._has_already_handler(signum):
            if not replace_existing:
                log.info(
                    f"Signal {signum.name} already has a handler. Adding ours to the existing one."
                )
                handlers.append(signal.getsignal(signum))
            else:
                log.info(
                    f"Replacing existing handler for signal {signum.name} with ours."
                )

        self._register_signal(signum, _HandlersCompose(handlers))
        log.info(f"Registered {len(handlers)} handlers for signal {signum.name}.")

    @override
    def register_signal_handlers(self) -> None:
        if not (auto_requeue_signals := self._auto_requeue_signals()):
            log.info(
                "No auto-requeue signals found. Reverting to default Lightning behavior."
            )
            return super().register_signal_handlers()

        self.received_sigterm = False
        self._original_handlers = self._get_current_signal_handlers()

        signals = defaultdict[signal.Signals, list[_HANDLER]](lambda: [])
        signals[signal.SIGTERM].append(self._sigterm_notifier_fn)

        environment = self.trainer._accelerator_connector.cluster_environment
        if isinstance(environment, SLURMEnvironment):
            log.info("SLURM auto-requeueing enabled. Setting signal handlers.")
            for signal_handler in auto_requeue_signals:
                signals[signal_handler].append(self._slurm_sigusr_handler_fn)

        if isinstance(environment, LSFEnvironment):
            # Important note from https://amrex-astro.github.io/workflow/olcf-workflow.html:
            # We can also ask the job manager to send a warning signal some amount of time before the allocation expires by passing -wa 'signal' and -wt '[hour:]minute' to bsub. We can then have bash create a dump_and_stop file when it receives the signal, which will tell Castro to output a checkpoint file and exit cleanly after it finishes the current timestep. An important detail that I couldn't find documented anywhere is that the job manager sends the signal to all the processes in the job, not just the submission script, and we have to use a signal that is ignored by default so Castro doesn't immediately crash upon receiving it. SIGCHLD, SIGURG, and SIGWINCH are the only signals that fit this requirement and of these, SIGURG is the least likely to be triggered by other events.

            log.info("LSF auto-requeueing enabled. Setting signal handlers.")
            for signal_handler in auto_requeue_signals:
                signals[signal_handler].append(self._lsf_sigusr_handler_fn)

        for signum, handlers in signals.items():
            if not handlers:
                continue

            self._compose_and_register(signum, handlers)

    def _should_ignore_signal_handler(self) -> str | None:
        if threading.current_thread() is not threading.main_thread():
            return "Not in main thread"

        if torch.utils.data.get_worker_info() is not None:
            return "Inside DataLoader worker process"

        return None

    @override
    def _slurm_sigusr_handler_fn(self, signum: _SIGNUM, _: FrameType) -> None:
        if ignore_reason := self._should_ignore_signal_handler():
            log.info(
                f"Skipping SLURM auto-requeue signal handler. Reason: {ignore_reason}"
            )
            return

        log.critical(f"Handling SLURM auto-requeue signal: {signum}")

        # save logger to make sure we get all the metrics
        for logger in self.trainer.loggers:
            logger.finalize("finished")

        hpc_save_path = self.trainer._checkpoint_connector.hpc_save_path(
            self.trainer.default_root_dir
        )
        self.trainer.save_checkpoint(hpc_save_path)

        if not self.trainer.is_global_zero:
            return

        # find job id
        array_job_id = os.getenv("SLURM_ARRAY_JOB_ID")
        if array_job_id is not None:
            array_task_id = os.environ["SLURM_ARRAY_TASK_ID"]
            job_id = f"{array_job_id}_{array_task_id}"
        else:
            job_id = os.environ["SLURM_JOB_ID"]

        assert re.match("[0-9_-]+", job_id)
        cmd = ["scontrol", "requeue", job_id]

        # requeue job
        log.info(f"Requeuing job {job_id}...")
        try:
            result = subprocess.call(cmd)
        except FileNotFoundError:
            # This can occur if a subprocess call to `scontrol` is run outside a shell context
            # Re-attempt call (now with shell context). If any error is raised, propagate to user.
            # When running a shell command, it should be passed as a single string.
            result = subprocess.call(" ".join(cmd), shell=True)

        # print result text
        if result == 0:
            log.info(f"Requeued SLURM job: {job_id}")
        else:
            log.warning(f"Requeuing SLURM job {job_id} failed with error code {result}")

    def _lsf_sigusr_handler_fn(self, signum: _SIGNUM, _: FrameType) -> None:
        if ignore_reason := self._should_ignore_signal_handler():
            log.info(
                f"Skipping LSF auto-requeue signal handler. Reason: {ignore_reason}"
            )
            return

        log.critical(f"Handling LSF auto-requeue signal: {signum}")

        # Save logger to make sure we get all the metrics
        for logger in self.trainer.loggers:
            logger.finalize("finished")

        # Save checkpoint
        hpc_save_path = self.trainer._checkpoint_connector.hpc_save_path(
            self.trainer.default_root_dir
        )
        self.trainer.save_checkpoint(hpc_save_path)
        log.info(f"Saved checkpoint to {hpc_save_path}")

        if not self.trainer.is_global_zero:
            return

        # Find job id
        if (job_id := os.getenv("LSB_JOBID")) is None:
            log.warning(
                "LSB_JOBID environment variable not found. Unable to requeue job."
            )
            return

        assert re.match("[0-9_-]+", job_id)

        exe = "brequeue"
        if (bin_dir := os.getenv("LSF_BINDIR")) is not None:
            exe = str((Path(bin_dir) / exe).resolve().absolute())

        log.info(f"Using LSF requeue executable: {exe}")

        # If NSHRUNNER_LSF_EXIT_SCRIPT_DIR exists, we should emit a bash script in that directory
        # rather than calling the requeue command directly. This is because the requeue command
        # is only available outside of the `jsrun` context, and the exit script is called within
        # the `jsrun` context.
        if not (exit_script_dir := os.getenv("NSHRUNNER_LSF_EXIT_SCRIPT_DIR")):
            cmd = [exe, job_id]

            # Requeue job
            log.info(f"Requeuing job {job_id}...")
            try:
                result = subprocess.call(cmd)
            except FileNotFoundError:
                # Retry with shell context if subprocess call fails
                result = subprocess.call(" ".join(cmd), shell=True)

            # Print result text
            if result == 0:
                log.info(f"Requeued LSF job: {job_id}")
            else:
                log.warning(
                    f"Requeuing LSF job {job_id} failed with error code {result}"
                )
        else:
            log.critical(
                "Environment variable NSHRUNNER_LSF_EXIT_SCRIPT_DIR found.\n"
                "Writing requeue script to exit script directory."
            )
            exit_script_dir = Path(exit_script_dir)
            assert (
                exit_script_dir.is_dir()
            ), f"Exit script directory {exit_script_dir} does not exist"

            exit_script_path = exit_script_dir / f"requeue_{job_id}.sh"
            log.info(f"Writing requeue script to {exit_script_path}")

            with exit_script_path.open("w") as f:
                f.write(f"#!/bin/bash\n{exe} {job_id}\n")

            # Make the script executable
            os.chmod(exit_script_path, 0o755)

            log.info(f"Requeue script written to {exit_script_path}")

            # Kill the current session to trigger the exit script
            log.info("Killing current session to trigger exit script")
            self._kill_current_session()

    def _kill_current_session(self):
        from lightning.pytorch.trainer.call import _interrupt

        _interrupt(self.trainer, KeyboardInterrupt())
        self.trainer._teardown()
        if (launcher := self.trainer.strategy.launcher) is not None:
            launcher.kill(_get_sigkill_signal())
        exit(1)


def _get_sigkill_signal() -> _SIGNUM:
    return signal.SIGTERM if sys.platform == "win32" else signal.SIGKILL
